library fields.field;

import 'dart:async';
import 'dart:mirrors';
import 'field_list.dart';
import 'mirrors.dart';

/// A marker class for field annotations.
abstract class FieldAnnotation {
  const FieldAnnotation();  
}

const IgnoreField = const _IgnoreField();
class _IgnoreField extends FieldAnnotation {
  const _IgnoreField();
}

class DisplayName extends FieldAnnotation {
  final String name;
  const DisplayName(this.name);
}

const ReadOnly = const _ReadOnly();
class _ReadOnly extends FieldAnnotation {
  const _ReadOnly();
}

class UseFieldEditor extends FieldAnnotation {
  final Type editorType;
  const UseFieldEditor(this.editorType);
}

class Field {
  
  static Map _parseMetadata(DeclarationMirror declaration) {
    assert(declaration != null);
    final metadata = new Map()
        ..[UseFieldEditor] = null
        ..[ReadOnly] = false
        ..[IgnoreField] = false
        ..[DisplayName] = null
        ..[FieldAnnotation] = <FieldAnnotation>[];
    declaration.metadata.forEach((InstanceMirror mirror) {
      final instance = mirror.reflectee;
      if (instance is UseFieldEditor) {
        // Disallow > 1 @UseFieldEditor
        assert(metadata[UseFieldEditor] == null);
        // TODO(rms): validate that the editor type isAssignableTo FieldEditor.
        metadata[UseFieldEditor] = reflectClass(instance.editorType);
      } else if (instance == ReadOnly) {
        metadata[ReadOnly] = true;
      } else if (instance == IgnoreField) {
        metadata[IgnoreField] = true;
      } else if (instance is DisplayName) {
        // Disallow > 1 @DisplayName
        assert(metadata[DisplayName] == null);
        metadata[DisplayName] = instance.name;
      } else if (instance is FieldAnnotation) {
        // Editor-specific annotations, store them in the field as-is.
        metadata[FieldAnnotation].add(instance);
      }
    });
    return metadata;
  }
  
  factory Field.forGetter(
      InstanceMirror owner, 
      Symbol name, 
     {MethodMirror getter,
      Field parent}) {
    if (getter == null) {
      getter = owner.type.getters[name];
      if (getter == null) {
        throw new ArgumentError(
            '$name is not a getter of ${owner.type.qualifiedName}');
      }
    }
    final setter = new Symbol('${MirrorSystem.getName(name)}=');
    final metadata = _parseMetadata(getter);      
    if (metadata[IgnoreField]) {
      return null;
    }
    return new Field._(
        owner, 
        name, 
        typeOf(getter.returnType), 
        parent: parent,
        editor: metadata[UseFieldEditor],
        annotations: metadata[FieldAnnotation],
        displayName: metadata[DisplayName],
        isPrivate: getter.isPrivate,
        isReadOnly: !owner.type.setters.containsKey(setter) 
          || metadata[ReadOnly]);
  }
  
  factory Field.forVariable(
      InstanceMirror owner, 
      Symbol name, 
     {VariableMirror variable,
      Field parent}) {
    if (variable == null) {
      variable = owner.type.variables[name];
      if (variable == null || variable.isStatic) {
        throw new ArgumentError(
            '$name is not an instance variable of ${owner.type.qualifiedName}');
      }
    }
    final metadata = _parseMetadata(variable);      
    if (metadata[IgnoreField]) {
      return null;
    }
    return new Field._(
        owner, 
        name, 
        typeOf(variable.type), 
        parent: parent,
        editor: metadata[UseFieldEditor],
        annotations: metadata[FieldAnnotation],
        displayName: metadata[DisplayName],
        isPrivate: variable.isPrivate,
        isReadOnly: variable.isFinal || metadata[ReadOnly]);
  }
  
  final Symbol name;
  final String displayName;
  final Type type;
  final bool isPrivate;
  final bool isReadOnly;
  final Field parent;
  final ClassMirror editor;
  final Iterable<FieldAnnotation> annotations;
  FieldList _children;
  FieldList get children => _children;
  InstanceMirror owner;  
  Object _value;
  
  final _values = new StreamController<Object>.broadcast();
  Stream<Object> get values => _values.stream;
  
  Field._(
      this.owner, 
      Symbol name, 
      this.type, 
     {this.parent, 
      this.editor, 
      this.annotations: const [], 
      String displayName, 
      this.isPrivate: false, 
      this.isReadOnly: false})
      : this.name = name
      , this.displayName = 
          (displayName == null) ? MirrorSystem.getName(name) : displayName {
    _children = new FieldList(this);
    _value = owner.getField(name).reflectee;
    if (!isSimpleType(type)) {
      _children.refreshSource(_value);
    }
  }
  
  Future<Object> getValue() {
    return owner.getFieldAsync(name)
        .then((instance) {
          if (instance.reflectee != _value) {
            _value = instance.reflectee;
            _refreshChildren().then((_) => _values.add(instance.reflectee));
          }
          return new Future.value(_value);
        })
        .catchError((e) => null,
              test: (e) => e is MirroredUncaughtExceptionError 
                        || e is MirroredCompilationError);
  }
  
  void refreshValue() { 
    getValue();
  }
  
  Future<Object> setValue(Object val) {
    if(isReadOnly) {
      return new Future.error(
          new StateError('Field $displayName is read-only.'));
    }
    else {
      if (val == _value) {
        return new Future.value(val);
      }
      else {
        return owner.setFieldAsync(name, reflect(val)).then((_) {
          _value = val;
          return _refreshChildren().then((_) {
            _values.add(val);
            if (parent != null) {
              parent._values.add(parent._value);
            }
            return new Future.value(val);
          });
        });
      }
    }
  }
      
  Future _refreshChildren() {    
    if (children.isEmpty) {
      return new Future.value();
    }
    children.refreshSource(_value);
    return Future.forEach(children, (f) => f.refreshValue());
  }
}
